<?php
/**
 * @file 
 * Additional per-module functionality used by import_html.
 * 
 * Each available module MAY impliment a hook_import_html() function to add or manipulate
 * its own data as the node gets saved.
 * 
 * @ingroup import_html Import HTML
 * @author Dan Morrison http://coders.co.nz/
 */
 
/**
 * hook_import_html - prototype function hook
 * 
 * invoke this when creating a node out of nothing
 * 
 * @param $profile array of settings for the import_html process
 * @param $node node object with lots of extra data added to it, only some of which we want to save.
 * @param $datadoc XML doc of the 'simple' node source that we may investigate for extra info if the cooked node object is not good enough 
 */ 
function hook_import_html($profile, &$node, $datadoc = NULL) {
  
} 
/**
 * hook_import_html_after_save - prototype function hook
 * 
 * Some functions (alias) can only operate when the nid is already known.
 * Invoke this when the node has been created
 * 
 * @param $profile array of settings for the import_html process
 * @param $node node object with lots of extra data added to it, only some of which we want to save.
 * @param $datadoc XML doc of the 'simple' node source that we may investigate for extra info if the cooked node object is not good enough 
 */ 
function hook_import_html_after_save($profile, &$node, $datadoc = NULL) {
  
} 



/**
 * Implementation of hook_import_html 
 * 
 * Add appropriate normal node attributes - title, body and teaser
 * - to the node
 */ 
function node_import_html($profile, &$node, $datadoc = NULL) {
  //  Title
  // Note, internally the title should always be plaintext, encoded characters and no newlines
  // A meta 'title' may have been set, 
  // but I'll use the real visible one by preference.
  $h1 = xml_get_element_content($datadoc, 'h1');
  if ($h1) {
    $node->title = $h1;
  } 
  else {
    $title_elem = xml_get_element_content($datadoc, 'title');
    if ($title_elem) {
      $node->title = $title_elem;
    }
  }

  //
  // BODY is the thing with id=content or whatever is set in the prefs
  // 
  debug_pre(array('XML DOM being scanned by node_import_html() for XPATH data extraction to find the body content' => xml_tostring($datadoc)), 3);
  $content_element = xml_getelementbyid($datadoc, 'content');  

  if (!$content_element) {
    drupal_set_message("Failed to find a body, anything with id='content' in this page");
    // TODO some cck may be valid without a 'content' body?
  }

  if ($content_element) {
    $node->body = trim(xml_tostring($content_element, TRUE));

    // The XSL may give us further hints about how it found this body data
    $precision = xml_getattribute($content_element, 'precision');
    if ($precision != NULL) {
      $node->import_html_precision = $precision;
      $explanation = xml_getattribute($content_element, 'title');
      debug(t("I'm %precision % confident that we found the right body element. !explanation", array('%precision' => 100*$precision, '!explanation' => $explanation)));
    }
  }

  // It's possible that our input was content-encoded (if it came from RSS or the old import-xml node)
  // If so, the entities should be unwrapped.
  // Other (HTML) imports should not require this
  // @TODO when was $node->content_encoded being flagged?
  $node->body = ($node->content_encoded) ? html_entity_decode($node->body):$node->body;

  //
  //  Teaser
  //
  $teaser = xml_textcontent( xml_getelementbyid($datadoc, 'teaser') ); 
  $description = xml_textcontent( xml_getelementbyid($datadoc, 'description') );
  if (!$teaser) $teaser = $description ;
  // No custom teaser found, but
  // Imported HTML is highly likely to contain formatting that
  // will ruin teasers. Do our own cut-down version that assumes 
  // FILTERED HTML
  $teaser = filter_filter('process', 0, 1, node_teaser($node->body));
  if ($teaser) $node->teaser = $teaser;
}


/**
 * Implementation of hook_import_html 
 * 
 * Add appropriate user identification to the new node.
 */ 
function user_import_html($profile, &$node, $datadoc = NULL) {
  global $user;
  $node->uid = $user->uid;
  $node->name = $profile['import_user'];
  if ($import_user = user_load(array('name' => $node->name))) {
    $node->uid = $import_user->uid;
  }
}

  
/**
 * Set alias and navigation for the given node
 * Helper for import_html_import_files()
 */
function menu_import_html_after_save($profile, &$node, $datadoc = NULL) {
  $normal_path = "node/{$node->nid}";
 
  // Create Menu item for this new item
  if ($profile['create_menus']) {
    $node->label = $node->title;

    // Need to beware of stupid long titles - they can't fit in breadcrumbs and menus
    if (strlen($node->title) > IMPORT_HTML_MAX_LABEL_LENGTH) {
      $path_bits = split('/', $node->path);
      $node->label=array_pop($path_bits);
      $node->label=preg_replace('/\?.*$/', '?', $node->label); // messiness from mirrored URLs with args in 
      $node->label=str_replace('_', ' ', $node->label);
      // @TODO maybe adjust this title-munging algoritm to make better guesses
      drupal_set_message(t("When creating menu item, the title '%node_title is too long to fit into a menu label (MAX=%MAX chars), truncating it to '%node_label' . This is not terrible, just unaesthetic. You may want to fix this later.", array('%node_title' => $node->title, '%MAX' => IMPORT_HTML_MAX_LABEL_LENGTH, '%node_label' => $node->label)));
    }

    // Check if a placeholder for the alias exists. 
    // This can happen if the child items were created earlier and required a placeholder
    // Try both canonic and aliased paths
    if( ! $menu = menu_get_item_by_path($normal_path) ) {
      $menu = menu_get_item_by_path($node->path);
    }
    if ($menu) {
      drupal_set_message(t("Discovered that a placeholder menu item for %path already exists. Filling that one out with the full details now", array('%path' => $node->path)));
    }
    else {
      // New menu item
      debug("Looking for parent menu of $node->path", 2);
      $pid = import_html_create_menu_path(dirname($node->path), '');
      if (!$pid) { $pid = $profile['menu_parent_id']; }
      $menu = array(
        'pid' => $pid,
        'type' => MENU_CREATED_BY_ADMIN | MENU_MODIFIABLE_BY_ADMIN | MENU_VISIBLE_IN_TREE | MENU_VISIBLE_IN_BREADCRUMB,
      );
    }
    $menu['title'] = $node->label;
    $menu['path'] = $normal_path;
    $node->menu = $menu;
    menu_edit_item_save($node->menu);
  }
}

/**
 * Given a traditional URL path, return the parent menu item
 *
 * A helper function to menu_import_html_after_save()
 * 
 * This is a recursive function that slices the given
 * (alias) path up until it finds a menu item it recognises.
 * Once found, it comes back down again building a path
 * until we end up with the new parent menu id.
 *
 * @param $path the path (aliased URL) we intend to construct
 * @param $title optional Display title of the menu item at that path point, otherwist it'll use the last fragment of the path as a display title
 * @return a menu id. FALSE if it hit the top unexpectedly
 * */
function import_html_create_menu_path($path, $title = '') {
  debug("Looking for a menu item matching path '$path' ", 3);
  if ((!module_exists('menu')) || (!module_exists('path'))) {
    return;
  }

  // check parent exists
  if ($pid = menu_get_mid_by_path($path)) {
    return $pid;
  }

  debug("Failed to find a menu item '$path', making it.", 2);

  $p_path = dirname($path); // parent path

  if ($p_path == dirname($p_path)) {
    return import_html_variable('menu_parent_id');
    /*
    debug("
        Looping menu path, this is either a recursion error (illegal paths),
        or caused by creating a new menu tree from the bottom up and we just
        reached the top.
        Although a menu structure for this page is now built, it may be hard
        to navigate to child pages if the parent doesn't exist.
        This can be fixed by importing the parents first or creating section header content.
    ", 2);
    return FALSE;
    */
  }

  if (!$pid = menu_get_mid_by_path($p_path)) {
    // Parent not found,
    // Need to make it up...
    $pid = import_html_create_menu_path($p_path);
    // That will have climbed up and come back down. with its nearest new ancestor
  }
  // and attach self to it

  if ($pid) {
    if (!$title) {
      $title = basename($path);
      if (strstr(".", $title)) {
        $title = substr($title, 0, strrpos($title, "."));
      }
    }
    $new = array(
      'title' => $title,
      'path' => $path,
      'pid' => $pid,
      'type' => MENU_CUSTOM_ITEM
    );
    debug("import_html_create_menu_path : Defining new menu item '$path' as a child of $pid", 3);

    if (!form_get_errors() && (!menu_get_mid_by_path($path))) {
      menu_edit_item_save($new);
    }
    else {
      debug_pre(form_get_errors());
      debug_pre(menu_get_menu(), 4);
      return;
    }
    $mid = menu_get_mid_by_path($path);
    
#    // need to clear the menu cache for the new item to show up
#    menu_rebuild();

    debug("New Menu Item '". $new['title'] ."' id is :". $mid, 2);
    return $mid;
  }
  else {
    debug("Failed to regressively build menu for '$path'. This should not happen. Dunno why we are here.");
  }
}


 
/**
 * Add path alias support to the import_html process
 * 
 * Adds an old-style (legacy) alias to the node path if required.
 */ 
function path_import_html_after_save($profile, &$node, $datadoc = NULL) {
  #dpm(array('INVOKED path_import_html_after_save for '. $node->path));
  if ($profile['legacy_aliases'] && ($node->old_path != $node->path)) {
    debug("Setting up navigation links for this item now. Directing {$node->old_path} to go to the system path ". drupal_get_normal_path($node->path), 2);

    $normal_path = "node/{$node->nid}";
    
    if ($normal_path == $node->path) {
      drupal_set_message(t("Failed to resolve %node_path into a system path. Cannot create alias at this time.", array('%node_path' => $node->path)));
    } 
    else {
      path_set_alias($normal_path, $node->old_path);
      drupal_set_message(t("This document (known internally as '%normal_path' ) should now be accessible via aliases as both '!main_alias' and '!legacy_alias' ",
        array(
          '%normal_path' => $normal_path,
          '!main_alias' => l($node->path, $node->path),
          '!legacy_alias' => l($node->old_path, $node->old_path)
        )
      ));
    }
  }
  
}


/**
 * Absorbs elements in the import document we recognise as being (probably)
 * taxonomy terms.
 * 
 * 
 * Any link rel="tag" is a start. The plaintext content of such a tag is
 * searched for as a term.
 * 
 * both <a href rel='tag' > and <link rel="tag"> are good for me.
 * 
 * @TODO this still needs to be expanded with real-world cases
 */
function taxonomy_import_html($profile, &$node, $datadoc = NULL) {

  $active_vocabs = taxonomy_get_vocabularies($node->type);
  if(!$active_vocabs) { return; }
  // Scan to see if any allow freetagging. After that it's guesswork which one to target. The one with the lowset number;
  $prime_vocab = 0;
  foreach($active_vocabs as $vocabulary) {
    if($vocabulary->tags) {
      $prime_vocabulary = $vocabulary;
      break;
    }
  }

  // Translate links with rel=tag into taxonomy terms
  $relationships = xml_query($datadoc, './/*[@rel]');
  foreach ($relationships as $link) {
    if (empty($link)) {continue;}
    $reltype = xml_getattribute($link, 'rel');
    // rels can be multiple - like classes. Split on space
    $reltypes = explode(' ', xml_getAttribute($link, 'rel'));
    $rellink = xml_getattribute($link, 'href');
    $label = trim(xml_textcontent($link));
    // Microformats may put their value into a 'title' attribute instead 
    if (empty($label)) {
      $label = xml_getAttribute($link, 'title'); 
    }

    if (!empty($label)) {
      foreach($reltypes as $reltype) {
        if ($reltype == 'tag') {
          // Some of this magic should probably shift into absorb_metas()
          // Still thinking of the best way to encode tags into the raw HTML
  
          if (strstr($label, ":")) {
            // allow an extended format for this value
            // to support freetext vocab additions
            // <a rel='tag'>Author:Joe Brown</a> will create a term in the 'Author' vocab called "Joe Brown".
            debug("Splitting rel link '$label' up to make it a vocab term entry", 2);
            $vocab_term = split(":", $label);
            $vocab = $vocab_term[0];
            $label = $vocab_term[1];
            $vocabulary = taxonomy_get_vocabulary_by_name($vocab);
            if ($vocabulary) {
              $node->taxonomy['tags'][$vocabulary->vid] .= ",". $label;
            }
          }
  
          // Match against existing terms
          $terms = taxonomy_get_term_by_name($label);
          // Allow an optional 'S' on the end of terms when looking for a match. Just to be flexible.
          $terms = array_merge($terms, taxonomy_get_term_by_name($label ."s"));
  
          if (!is_array($node->taxonomy)) {$node->taxonomy = array();}
          if ($terms) {
            foreach($terms as $term) {
              // If we did successfully identify the vocab earlier, filter on that. Otherwise let anything go
              if ($vocabulary && ($term->vid != $vocabulary->vid)) { continue; }
              $node->taxonomy[$term->tid] = $term;
            }
          }
          else {
            // Couldn't retrieve it, can we add freetags?
            $node->taxonomy['tags'][$prime_vocabulary->vid] .= ",". $label;
          }

        } // is a 'tag'
        import_html_absorb_properties($node, $reltype, $label);
      }  // inner rel tag type
    } // has value
  } // each rel
}



/**
 * Utility Functions which should exist but don't yet.
 * out-of-module names sorry.
 */


/**
 * Return a menu item matching a given path or alias.
 * 
 * Menu_get_item re-writes the path from alias to 'normal' path when it
 * initializes. This makes it impossible to retrieve menu items by name using the available API.
 * 
 * @return a menu ID
 */
function menu_get_mid_by_path($path) {
  // Caching works against me when adding lots of aliases in one go. Wipe it :(
  drupal_lookup_path('wipe'); // This is not working?
  $alias = db_result(db_query("SELECT src FROM {url_alias} WHERE dst = '%s'", $path));

  $menu = menu_get_menu();
  if ($mid = $menu['path index'][$path]) {
    return $mid;
  }
  if ($alias && $mid = $menu['path index'][$alias]) {
    return $mid;
  }

  // it's bad mojo to mess with the DB directly, but menu doesn't provide a lookup API.
  // Or a way to avoid caching. Do it by hand if I need a newly added menu
  $row = db_fetch_object(db_query('SELECT * FROM {menu} WHERE path != "" AND (path = "%s" OR path = "%s") ', $path, $alias ));
  if ($row) {
    return $row->mid;
  }
}
/**
 * Return a menu item matching a given path or alias.
 */
function menu_get_item_by_path($path) {
  $result = db_query('SELECT * FROM {menu} WHERE path="%s"', $path);
  return db_fetch_array($result);
}


function taxonomy_get_vocabulary_by_name($name) {
  $vs = taxonomy_get_vocabularies();
  foreach ($vs as $voc) {
    if ($voc->name == $name) return $voc;
  }
}
